#!/usr/bin/env python3

import argparse
import asyncio
import gettext
import logging
import pdb
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, FileType
from datetime import datetime, time, timedelta
from functools import partial

import colorama as ca

from impetuous.cli import parse_friendly_datetime
from impetuous.cli.actions import (FUNCTION_DEST, Cli, DryRun, im_a,
                                   im_config_edit, im_decode, im_doing,
                                   im_edit, im_encode, im_export, im_import,
                                   im_import_legacy, im_post, im_repl, im_show,
                                   im_summary, im_suggest)
from impetuous.im import Impetuous
from impetuous.interaction import OperationError, RequestError
from impetuous.sheet import localtz, utcnow

logger = logging.getLogger('impetuous')


def main():
    # Localization? Maybe?
    import os.path
    ld = os.path.join(os.path.dirname(__file__), '..', 'locale')
    gettext.install('impetuous', ld)

    today = utcnow().astimezone(localtz).date()
    since = datetime.combine(today, time(0))
    until = since + timedelta(days=1)

    # Some arguments we want to know earlier, but we don't want
    # parse_known_args to throw up if it sees '-h', so we create a parser with
    # add_help=False, then add help later on after parsing early arguments.
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter, add_help=False)
    parser.add_argument('--dbecho', action='store_true', default=False,
                        help='Show everything that sqlalchemy is doing. You almost never want to do this.')
    parser.add_argument('--log', help='Logging level',
                        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR'],
                        default='WARNING')
    parser.add_argument('--since', help='Operate on entries starting after ...',
                        type=parse_friendly_datetime, default=since)
    parser.add_argument('--until', help='Operate on entries ending before ...',
                        type=parse_friendly_datetime, default=until)
    parser.add_argument('-y', '--yesterday', help='Operate on yesterday. Can be repeated.',
                        action='count', default=0)
    parser.add_argument('-d', '--debug', action='store_true', default=False)
    args, _ = parser.parse_known_args()

    # Initialize colorama stuff
    ca.init()

    logging.basicConfig(level=getattr(logging, args.log))
    utcnow_ish = utcnow() - timedelta(days=args.yesterday)
    datetime_arg = partial(parse_friendly_datetime, default=utcnow_ish)

    try:
        impetuous = Impetuous(dbecho=args.dbecho)
        cli = Cli(impetuous)
    except KeyboardInterrupt:
        logger.info("Bye!")
    except SystemExit:
        raise
    except Exception as e:
        if args.debug:
            logger.exception(e)
            pdb.post_mortem()
        raise

    parser.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS)

    subparser = parser.add_subparsers(title='action')
    subparser.required = True
    subparser.dest = FUNCTION_DEST

    show_help = '''Show a list things done, startings, endings, and durations
    for the current sheet.'''
    show_args = subparser.add_parser('show', help=show_help, aliases=['s', 'l'], formatter_class=ArgumentDefaultsHelpFormatter)
    show_args.add_argument('--reverse', '-r', action='store_true', default=False)
    show_args.add_argument('--verbose', '-v', action='store_true', default=False)
    show_args.add_argument('-l', '--limit', type=int, default=None)
    show_args.set_defaults(action=im_show)

    summary_help = '''Show summary organized.'''
    summary_args = subparser.add_parser('summary', help=summary_help, aliases=['ss', 'll'], formatter_class=ArgumentDefaultsHelpFormatter)
    summary_args .add_argument('--verbose', '-v', action='store_true', default=False)
    summary_args.set_defaults(action=im_summary)

    doing_help = '''Try to stop the thing running soonest before the given
    stopping time (now by default) and then start something new if you give me
    "blah".'''
    doing_args = subparser.add_parser('doing', help=doing_help, aliases=['d', 'doing'], formatter_class=ArgumentDefaultsHelpFormatter)
    doing_args.add_argument('blah', type=str, nargs='*')
    doing_args.add_argument('--when', '-w', type=datetime_arg, default='now', help='Start of period')
    doing_args.add_argument('--comment', '-c', type=str)
    doing_args.add_argument('-n', '--dry_run', action='store_true', default=False, help="Don't commit changes to the database. Just talk about them.")
    doing_args.set_defaults(action=im_doing)

    edit_help = '''Opens a YAML representation of a bunch of time entries in a
    temporary file to be modified in EDITOR.'''
    edit_args = subparser.add_parser('edit', help=edit_help, formatter_class=ArgumentDefaultsHelpFormatter)
    edit_args.add_argument('-n', '--dry_run', action='store_true', default=False, help="Don't commit changes to the database. Just talk about them.")
    edit_args.set_defaults(action=im_edit)

    suggest_help = '''Suggest something to do, like if you're bored.'''
    suggest_args = subparser.add_parser('suggest', help=suggest_help, formatter_class=ArgumentDefaultsHelpFormatter)
    suggest_args.add_argument('-l', '--limit', type=int, default=20, help="How many suggestions.")
    suggest_args.set_defaults(action=im_suggest)

    import_legacy_help = '''Import legacy/v1 time sheet. Run im -l=INFO for
    more output. Run import-legacy -q for less. Nine out of ten cat herders
    recommend running this with ~/.config/impetuous/sheets/*
    ~/.local/share/impetuous/* and then making your local impetuous maintainer
    a pot of coffee.
    This does not respect --since or --until.'''
    import_legacy_args = subparser.add_parser('import-legacy', help=import_legacy_help, formatter_class=ArgumentDefaultsHelpFormatter)
    import_legacy_args.add_argument('-n', '--dry_run', action='store_true', default=False, help="Don't commit changes to the database. Just talk about them.")
    import_legacy_args.add_argument('-q', '--quiet', action='store_true', default=False, help="Don't output the entries as they are imported.")
    import_legacy_args.add_argument('--ignore-errors', action='store_true', default=False, help="Commit partial data, even if some stuff couldn't be inserted.")
    import_legacy_args.add_argument('files', default='-', nargs='+', type=argparse.FileType('r'))
    import_legacy_args.set_defaults(action=im_import_legacy)

    config_help = '''Opens the config in EDITOR.'''
    config_args = subparser.add_parser('config-edit', help=config_help, formatter_class=ArgumentDefaultsHelpFormatter)
    config_args.set_defaults(action=im_config_edit)

    post_help = '''Submits workshows to JIRA and Freshdesk.'''
    post_args = subparser.add_parser('post', help=post_help, formatter_class=ArgumentDefaultsHelpFormatter)
    post_args.add_argument('-n', '--dry_run', action='store_true', default=False, help="Do not actually send submissions. Just talk about them.")
    post_args.set_defaults(action=im_post)

    encode_help = '''Encodes passwords/secrets in the config using a given
    codec. The config parser is very rude and will does not preserve comments
    anything in your config file when this modifies it.'''
    encode_args = subparser.add_parser('encode', help=encode_help, formatter_class=ArgumentDefaultsHelpFormatter)
    encode_args.add_argument('codec', choices=['base64', 'bz2', 'hex', 'quotedprintable', 'uu', 'zlib'])
    encode_args.set_defaults(action=im_encode)

    decode_help = '''Decodes passwords in the config using a given codec. The
    config parser is very rude and will does not preserve comments anything in
    your config file when this modifies it.'''
    decode_args = subparser.add_parser('decode', help=decode_help, formatter_class=ArgumentDefaultsHelpFormatter)
    decode_args.set_defaults(action=im_decode)

    a_help = '''What are you?'''
    a_args = subparser.add_parser('a', aliases=['an'], help=a_help, formatter_class=ArgumentDefaultsHelpFormatter)
    a_args.add_argument('?', type=str, nargs='+')
    a_args.set_defaults(action=im_a)

    repl_args = subparser.add_parser('repl', formatter_class=ArgumentDefaultsHelpFormatter)
    repl_args.set_defaults(action=im_repl)

    try:
        args = parser.parse_args()
    except KeyboardInterrupt:
        logger.info("Bye!")
    except SystemExit:
        raise
    except Exception as e:
        if args.debug:
            logger.exception(e)
            pdb.post_mortem()
        raise
    args.since -= timedelta(days=args.yesterday)
    args.until -= timedelta(days=args.yesterday)
    action = args.action

    logger.debug("Arguments are %s", args)
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(action(cli, args))
    except DryRun as e:
        logger.info("Exited because dry run!")
    except RequestError as e:
        logger.error(e)
        logger.error("Tried executing: %s", e.orig.statement)
        logger.error("With: %r", e.orig.params)
        raise SystemExit(1)
    except (ValueError, OperationError) as e:
        if args.debug:
            logger.exception(e)
            pdb.post_mortem()
        else:
            logger.error(e)
        raise SystemExit(1)
    except NotImplementedError as e:
        if args.debug:
            logger.exception(e)
            pdb.post_mortem()
        else:
            logger.error("This isn't implemented: %s", e)
        raise SystemExit(1)
    except KeyboardInterrupt:
        logger.info("Bye!")
    except SystemExit:
        raise
    except Exception as e:
        if args.debug:
            logger.exception(e)
            pdb.post_mortem()
        raise
    finally:
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()


if __name__ == '__main__':
    main()



""" Request Structures, the Python API to requests and the data and stuff
"""
import logging
from uuid import UUID

import attr
from attr.validators import instance_of, optional
from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql import and_, expression, operators

import impetuous.data

logger = logging.getLogger(__name__)


# TODO break this up into request structures, exceptions, other stuff?

# Validation errors the database driver will send us ...
error_messages = {"entry start end overlap": "This operation tries to make multiple time entries overlap."}


class RequestError(ValueError):
    def __init__(self, orig, *args, **kwargs):
        self.orig = orig
        super().__init__(*args, **kwargs)


class AccessProhibited(Exception):
    pass


class OperationError(Exception):
    pass


class TooFewResults(OperationError):
    pass


class TooManyResults(OperationError):
    pass


#@attr.s()
#class TableAccess(object):
#    tables = attr.ib()
#    where = attr.ib()


@attr.s()
class Agent(object):
    """ Represents a requester...
    """
    #def is_authenticated # TODO use ABC?


@attr.s()
class AccessControl(object):
    """
    Enforce your permissions here.

    All modification involves a table, columns and their values, and sometimes
    an existing row.
    This is enforced by access_to_modify().
    """
    tables = attr.ib()
    joins = attr.ib()
    # can create in resource (maybe if field values are a particular way)
    # can modify/delete a particular resource

    # all this will have to be redone probably or extended to allow for greater
    # querying... such as what resources can I access, what fields on them? what
    # conditions apply or filters are added implicitly on queries concerning
    # fields?

    # Interface might be...
    # On a resource/table; access to insert/view/delete some given entity/row.
    # On a entity/row; access to view/update given some field/column.
    # ... and access to use a join thiny ...

    # - Access to view/modify particular rows on a table should be supplimented
    # with a queryset that limits the rows it can operate on.
    # - It should also be possible to reject requests based on the fields. So
    # give users view/modify permissions to resource columns.

    def access_table_to_modify(self, agent, table_name, data):
        """
        This function should test if the agent can access the table, what rows
        they can access expressed by the where clause, and if they can alter the
        rows accordingly.

        Prohibited table and field value changes can be prevented by raising
        AccessProhibited.
        """
        if agent.is_authenticated():
            table = self.tables[table_name]
            for field in (*table.primary_key.columns.keys(), 'rev'):
                if field in data:
                    raise AccessProhibited(field)
            return table, None
        else:
            raise AccessProhibited()

    def access_table_to_read(self, agent, table_name, data):
        if agent.is_authenticated():
            table = self.tables[table_name]
            return table, None
        else:
            raise AccessProhibited()

    # Also aggregate things like fields to <verb> and ask for them all at once.

    def get_table(self, table_name):
        return self.tables[table_name]

    def access_table_field(self, agent, table_name, field_name):
        return self.get_table(table_name).c[field_name]

    def access_query_field(self, agent, query, field_name):
        return query.c[field_name]

    def access_join(self, agent, join_name):
        return impetuous.data.joins[join_name]

    def render_to_sql(self, obj, agent):
        """
        Given some request thing, return a sqlalchemy sql expression object if
        the agent can access the whatever.
        """
        a = AccessAttempt(self, agent)
        return a.render(obj)


impetuous_access_control = AccessControl(
    tables=impetuous.data.tables,
    joins=impetuous.data.joins,
)


@attr.s()
class AccessAttempt(object):
    """
    ac is the AccessControl
    agent is who is doing the accessing
    memory is what objects we've already rendered and the results (memoizing)
    working is what objects we're currently rendering

    TODO memory could introduce security badnesses if maybe some object is
    legal in one context but not the other?

    TODO capture access requests and do them all at the end in one thing...
    """
    ac = attr.ib()
    agent = attr.ib()
    memory = attr.ib(default=attr.Factory(dict))
    working = attr.ib(default=attr.Factory(set))
    #access = attr.ib(default=attr.Factory(list))

    def hash(self, obj):
        """
        We cheat on hashing all throughout here because:
          a) We only care about python object instance memorizing
          b) We don't care about object retrieval for `working`
          c) We don't play by the rules
        """
        return id(obj)

    def render(self, obj):
        obj_hash = self.hash(obj)
        assert obj_hash not in self.working, "Recursion bad!"
        self.working.add(obj_hash)
        try:

            if isinstance(obj, Comparison):
                sql = self._render_comparison(obj)
            elif isinstance(obj, Conjunction):
                sql = self._render_conjunction(obj)
            elif isinstance(obj, Gather):
                sql = self._render_gather(obj)
            elif isinstance(obj, Param):
                sql = self._render_param(obj)
            elif isinstance(obj, Field):
                sql = self._render_field(obj)
            elif isinstance(obj, Sort):
                sql = self._render_sort(obj)
            elif obj is None:  # Hrm..... this should probably not be allowed ...
                sql = None
            elif isinstance(obj, list):  # FIXME
                sql = obj
            else:
                # TODO FIXME decide on something here ...
                #sql = expression.literal(obj)
                raise NotImplementedError(f"Not sure what to do with {obj!r}.")

            self.memory[obj_hash] = sql
            return sql
        finally:
            self.working.remove(obj_hash)

    def _render_conjunction(self, obj):
        return obj.op(*(self.render(part) for part in obj.parts))

    def _render_comparison(self, obj):
        opfn = {
            'like': operators.like_op,
            'ilike': operators.ilike_op,
            'in': operators.in_op,
            'eq': operators.eq,
            'ne': operators.ne,
            'ge': operators.ge,
            'gt': operators.gt,
            'le': operators.le,
            'lt': operators.lt,
            'is': operators.is_,
            'is not': operators.isnot,
        }[obj.op]
        return opfn(self.render(obj.lh), self.render(obj.rh))

    def _render_param(self, obj):
        return expression.bindparam(obj.key, expanding=obj.expanding)

    def _render_gather(self, obj):
        where = None if obj.match is None else self.render(obj.match)
        # TODO maybe use render field here maybe? fix dis
        fields = [self.render(r) for r in obj.fields]
        # ... and here FIXME TODO XXX
        if obj.order_by:
            order_by = [self.render(r) for r in obj.order_by]
        else:
            order_by = None
        stmt = expression.select(fields, whereclause=where, order_by=order_by,
                                 limit=obj.limit)
        for join in obj.relating:
            stmt = stmt.select_from(self.ac.access_join(self.agent, join))
        return stmt

    def _render_fields(self, obj):
        """ Special handling for the fields part of the Gather
        """
        raise NotImplementedError

    def _render_field(self, obj):
        if obj.is_concrete:
            col = self.ac.access_table_field(self.agent, obj.at, obj.name)
        else:
            query = self.render(obj.at)
            col = self.ac.access_query_field(self.agent, query, obj.name)
        return col if obj.alias is None else col.label(obj.alias)

    def _render_sort(self, obj):
        # TODO this should be allowed XXX FIXME TODO lasjdfklasdjfklasdjfklaj
        #self._render_field(obj.field)
        if obj.order == 'asc':
            return self.render(obj.field).asc()
        elif obj.order == 'desc':
            return self.render(obj.field).desc()
        else:
            raise NotImplementedError(obj.sort)


@attr.s(frozen=True)
class Comparison(object):
    lh = attr.ib()
    op = attr.ib(validator=instance_of(str))
    rh = attr.ib()

    #def __or__(self, lh, rh):
    #    return Conjunction.or_(lh, rh)

    #def __and__(self, lh, rh):
    #    return Conjunction.and_(lh, rh)


@attr.s(frozen=True)
class Conjunction(object):
    # Op is actually as sqlalchemy expression, the only place where we only
    # directly reference sqlalchemy in the request stuff FIXME
    op = attr.ib()#validator=attr.validators.in_(['and', 'or']))
    # TODO, stronger validation, check the contents of the tuple ...
    parts = attr.ib(validator=instance_of(tuple),
                    converter=tuple)

    @classmethod
    def and_(cls, *parts):
        return cls(expression.and_, parts=parts)

    @classmethod
    def or_(cls, *parts):
        return cls(expression.or_, parts=parts)

    def __or__(self, lh, rh):
        return Conjunction.or_(lh, rh)

    def __and__(self, lh, rh):
        return Conjunction.and_(lh, rh)


@attr.s(frozen=True)
class Gather(object):
    # TODO try to remove converters, it messes with validation/error handling?
    fields = attr.ib(
        converter=tuple,
    )  # This has a validator down below
    # Verbing the attributes here is dumb... TODO XXX FIXME
    relating = attr.ib(
        validator=instance_of(tuple),
        converter=tuple,
        default=attr.Factory(tuple),
    )
    match = attr.ib(
        validator=optional(instance_of((Conjunction, Comparison))),
        default=None,
    )
    order_by = attr.ib(
        converter=tuple,
        default=attr.Factory(tuple),
    )
    limit = attr.ib(
        validator=optional(instance_of(int)),
        default=None,
    )

    @fields.validator
    def fields_tuple_of_fields(self, attribute, value):
        if not (isinstance(value, tuple) and all(isinstance(e, Field) for e in value)):
            raise ValueError(f"'{attribute.name}' must be a collection of fields.")

    @order_by.validator
    def order_by_tuple_of_sorts(self, attribute, value):
        if isinstance(value, tuple) and all(isinstance(e, Sort) for e in value):
            pass
        else:
            raise ValueError(f"'{attribute.name}' must be a collection of sorts.")

    #@fields.validator
    #def fields_dict_of_fields(self, attribute, value):
    #    if isinstance(value, dict)\
    #            and all(isinstance(k, str) and (isinstance(v, Field) or self.fields_dict_of_fields(attribute, v))
    #                    for k, v in value.items()):
    #        return "hip hip, horray!"
    #    else:
    #        raise ValueError(f"'{attribute.name}' must be a mapping of string to fields; but instead it's {value!r}.")


@attr.s(frozen=True)
class Param(object):
    key = attr.ib(validator=instance_of(str))
    expanding = attr.ib(validator=instance_of(bool), default=False)


@attr.s(frozen=True)
class Field(object):
    """
    A concrete field is a field on a resource, rather than a gathering. The
    resource is specified with a string.
    """
    at = attr.ib(validator=instance_of((str, Gather)))
    name = attr.ib(validator=instance_of(str))
    alias = attr.ib(default=None)

    @property
    def is_concrete(self):
        return isinstance(self.at, str)


@attr.s(frozen=True)
class Sort(object):
    order = attr.ib(validator=attr.validators.in_(('asc', 'desc')))
    field = attr.ib(validator=instance_of(Field))


# Requests

@attr.s(frozen=True)
class FindRequest(object):
    gather = attr.ib(validator=instance_of(Gather))
    using = attr.ib(validator=optional(instance_of(dict)),
                    default=attr.Factory(dict))


@attr.s(frozen=True)
class InsertRequest(object):
    resource = attr.ib(validator=instance_of(str))
    values = attr.ib()
    #fields = attr.ib()


@attr.s(frozen=True)
class UpdateRequest(object):
    resource = attr.ib(validator=instance_of(str))
    values = attr.ib()
    #match = attr.ib(validator=instance_of((Conjunction, Comparison)))
    id = attr.ib(validator=instance_of(UUID))
    rev = attr.ib(validator=optional(instance_of(UUID)), default=None)
    #fields = attr.ib()


@attr.s(frozen=True)
class DeleteRequest(object):
    resource = attr.ib(validator=instance_of(str))
    match = attr.ib(validator=instance_of((Conjunction, Comparison)))


def RequestError_from_IntegrityError(e):
    logger.debug("Derp? %r", e.orig.args)
    code, = e.orig.args
    msg = error_messages[code]
    raise RequestError(e, msg)


@attr.s()
class Interaction(object):
    """ A context around handling one or more requests.
    """

    conn = attr.ib()
    agent = attr.ib()
    ac = attr.ib(validator=instance_of(AccessControl))

    # TODO, maybe move those values items crap somewhere else
    def find(self, gather, using, map_values=None, unpack_values=None, map_items=None, unpack_items=None):
        # TODO enforce access stuff properly
        sql = self.ac.render_to_sql(gather, agent=self.agent)
        res = self.conn.execute(sql, **using)
        if map_values is not None:
            return (map_values(row.values()) for row in res)
        elif unpack_values is not None:
            return (unpack_values(*row.values()) for row in res)
        elif map_items is not None:
            return (map_items(row.items()) for row in res)
        elif unpack_items is not None:
            return (unpack_items(**dict(row.items())) for row in res)
        else:
            return res

    def find_one(self, *args, **kwargs):
        """ Raises TooFewResults or TooManyResults. Gather with limit=1 maybe.
        """
        res = self.find(*args, **kwargs)
        try:
            one = next(res)
        except StopIteration:
            raise TooFewResults("Expected one result, got zero.")
        try:
            next(res)
        except StopIteration:
            return one
        else:
            raise TooManyResults("Expected one result, got more than that.")
        

    def insert(self, resource, values):
        table, condition = self.ac.access_table_to_modify(self.agent, resource, values)
        assert condition is None, "what do we do here?"
        sql = expression.insert(table).values(**values)
        try:
            res = self.conn.execute(sql)
        except IntegrityError as e:
            raise RequestError_from_IntegrityError(e)
        return dict(zip(
            table.primary_key.columns.keys(),
            res.inserted_primary_key,
        ))

    def update(self, resource, values, id, rev=None):
        table, condition = self.ac.access_table_to_modify(self.agent, resource, values)
        where = table.c.id == id
        if rev is not None:
            where &= table.c.rev == rev
        if condition is not None:
            where &= condition
        sql = expression.update(table, whereclause=where).values(**values)
        try:
            res = self.conn.execute(sql)
        except IntegrityError as e:
            # An integrity error is probably a client value error...
            raise RequestError_from_IntegrityError(e)
        else:
            # TODO since the parameters are not where clauses but id and rev,
            # this should expect to update exactly one thing. This should
            # probably be named update_one or something ...
            if res.rowcount != 1:
                raise OperationError("I was supposed to update something but I couldn't find it.")
            #return res.rowcount

    def delete(self, resource, id, rev=None):
        table, condition = self.ac.access_table_to_modify(self.agent, resource, {})
        where = table.c.id == id
        if rev is not None:
            where &= table.c.rev == rev
        if condition is not None:
            where &= condition
        sql = expression.delete(table, whereclause=where)
        try:
            res = self.conn.execute(sql)
        except IntegrityError as e:
            raise RequestError_from_IntegrityError(e)
        else:
            if res.rowcount != 1:
                raise OperationError("I was supposed to delete something but I couldn't find it.")
            #return res.rowcount



"""empty message

Revision ID: f4aaf1bdf8c0
Revises: 
Create Date: 2018-03-17 00:07:16.470038

"""
from alembic import op
import sqlalchemy as sa

import impetuous.data

# revision identifiers, used by Alembic.
revision = 'f4aaf1bdf8c0'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('entry',
    sa.Column('id', impetuous.data.UUID(), nullable=False),
    sa.Column('rev', impetuous.data.UUID(), nullable=False),
    sa.Column('text', sa.String(), nullable=False),
    sa.Column('comment', sa.String(), nullable=False),
    sa.Column('start', impetuous.data.UTCDateTime(), nullable=False),
    sa.Column('end', impetuous.data.UTCDateTime(), nullable=True),
    sa.CheckConstraint('end > start', name='ends after start'),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index('idx_entry_start', 'entry', ['start'], unique=True)
    op.create_table('submission',
    sa.Column('id', impetuous.data.UUID(), nullable=False),
    sa.Column('rev', impetuous.data.UUID(), nullable=False),
    sa.Column('entry', impetuous.data.UUID(), nullable=False),
    sa.Column('ext', sa.String(), nullable=False),
    sa.Column('key', sa.String(), nullable=False),
    sa.Column('result', impetuous.data.YamlDoc(), nullable=False),
    sa.ForeignKeyConstraint(['entry'], ['entry.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index('idx_submission_entry_ext', 'submission', ['entry', 'ext', 'key'], unique=True)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('idx_submission_entry_ext', table_name='submission')
    op.drop_table('submission')
    op.drop_index('idx_entry_start', table_name='entry')
    op.drop_table('entry')
    # ### end Alembic commands ###



""" TODO rename this to models.py or something boring like that
"""

import datetime
import logging
import math
import os
import random
import re
import warnings
from uuid import UUID

import attr
import dateutil.parser
import yaml

from impetuous.serialization import UUID_PATTERN

logger = logging.getLogger(__name__)

utctz = datetime.timezone.utc
localtz = dateutil.tz.tzlocal()

not_loaded = type('not_loaded', (object,), {})()


def utcnow():
    if 'IM_NOW' in os.environ:
        dt = dateutil.parser.parse(os.environ['IM_NOW'])
        if dt.tzinfo is None:
            dt.replace(tzinfo=localtz)
        return dt.astimezone(datetime.timezone.utc)
    else:
        return datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)


class EntryNotFoundError(RuntimeError):  # TODO remove this
    pass


class ValidationError(Exception):

    def __init__(self, message, params=None):
        if params is None:
            params = {}
            super().__init__(message)
        else:
            super().__init__(message.format(**params))
        self.message = message
        self.params = params


@attr.s(cmp=True)
class Entry(object):
    id = attr.ib(default=None)
    rev = attr.ib(default=None)
    text = attr.ib(default='')
    comment = attr.ib(default='')
    start = attr.ib(default=None)
    end = attr.ib(default=None)
    submissions = attr.ib(default=attr.Factory(dict))

    @end.validator
    @start.validator
    def has_timezone(self, attribute, value):
        if value is None or value is not_loaded:
            pass
        elif value.tzinfo is None:
            raise ValueError("%s requires a timezone aware datetime, not %r" % (attribute.name, value))

    @classmethod
    def joker(cls, id=not_loaded, rev=not_loaded, text=not_loaded, comment=not_loaded, start=not_loaded, end=not_loaded, submissions=not_loaded):
        return cls(id=id, rev=rev, text=text, comment=comment, start=start, end=end, submissions=submissions)

    @classmethod
    def new(cls):
        return cls(id=None, rev=None, text='', comment='', start=None, end=None, submissions={})

    def __contains__(self, other):
        return (
                (other.end is not None and other.start <= self.start < other.end)
                or (self.end is not None and self.start <= other.start < self.end)
                )

    @property
    def duration(self):
        assert self.start is not not_loaded
        assert self.end is not not_loaded
        if self.end is None:
            end = utcnow()
        else:
            end = self.end
        return end - self.start

    @property
    def duration_in_minutes(self):
        """ Return duration in minutes as integer. Rounds randomly. :)
        """
        minutes_spent = self.duration.total_seconds() / 60
        if minutes_spent % 1 <= random.random():
            return math.floor(minutes_spent)
        else:
            return math.ceil(minutes_spent)


@attr.s(cmp=True)
class Submission(object):
    id = attr.ib(default=None)
    rev = attr.ib(default=None)
    entry = attr.ib(default=None)
    ext = attr.ib(default=None)
    key = attr.ib(default=None)
    result = attr.ib(default=None)

    @classmethod
    def joker(cls, id=not_loaded, rev=not_loaded, entry=not_loaded, ext=not_loaded, key=not_loaded, result=not_loaded):
        return cls(id=id, rev=rev, entry=entry, ext=ext, key=key, result=result)

    @property
    def entry_index(self):
        assert self.ext is not not_loaded  # assert double negation
        assert self.key is not not_loaded  # is not not readable...
        return self.ext, self.key


def load_cli_entry_from_yaml(loader, node):
    args = loader.construct_mapping(node, deep=True)
    if 'covert' in args:
        del args['covert']
    args['text'] = args.pop('task')
    for key in ('start', 'end'):
        if key in args:
            args[key] = args[key].astimezone(utctz)
    if 'submissions' in args:
        args['submissions'] = {s.entry_index: s for s in args['submissions']}
    return Entry(**args)


# I think this stuff is used by im_edit, should probably live in impetuous/cli/
# somewhere TODO

def load_cli_submission_from_yaml(loader, node):
    args = loader.construct_mapping(node, deep=True)
    return Submission(**args)


# http://stackoverflow.com/a/13295663
def timestamp_constructor(loader, node):
    return dateutil.parser.parse(node.value)


def load_UUID(loader, node):
    return UUID(loader.construct_scalar(node))


class cli_yaml_Loader(yaml.SafeLoader):
    pass

cli_yaml_Loader.add_constructor(u'tag:yaml.org,2002:timestamp', timestamp_constructor)
cli_yaml_Loader.add_constructor('!uuid', load_UUID)
cli_yaml_Loader.add_constructor(u'!entry', load_cli_entry_from_yaml)
cli_yaml_Loader.add_constructor(u'!submission', load_cli_submission_from_yaml)
cli_yaml_Loader.add_implicit_resolver('!uuid', UUID_PATTERN, None)


def dump_cli_entry_as_yaml(dumper, entry):
    as_dict = {
        'id': entry.id,
        'rev': entry.rev,
        'start': entry.start.astimezone(localtz),
        'task': entry.text,
    }
    if entry.end is not None:
        as_dict['end'] = entry.end.astimezone(localtz)
    if entry.submissions:
        as_dict['submissions'] = list(entry.submissions.values())
    if entry.comment:
        as_dict['comment'] = entry.comment
    return dumper.represent_mapping(u'!entry', as_dict)


def dump_cli_submission_as_yaml(dumper, result):
    return dumper.represent_mapping(u'!submission', attr.asdict(result))


def dump_UUID(dumper, uuid):
    return dumper.represent_scalar('!uuid', str(uuid))


class cli_yaml_Dumper(yaml.SafeDumper):
    pass

cli_yaml_Dumper.add_representer(UUID, dump_UUID)
cli_yaml_Dumper.add_representer(Entry, dump_cli_entry_as_yaml)
cli_yaml_Dumper.add_representer(Submission, dump_cli_submission_as_yaml)



# TODO move this to impetuous/cli/...
from datetime import datetime, timedelta
import re

import attr
import yaml
try:
    from yaml import CSafeLoader as yaml_Loader
except ImportError:
    from yaml import SafeLoader as yaml_Loader

from impetuous.sheet import Entry, Submission, timestamp_constructor, utctz


JIRA_PATTERN = re.compile(r'\w+-\d+')


def format_comments(comments):
    if len(comments) == 1:
        return comments[0]
    else:
        return '\n'.join('- %s' % comment for comment in comments)


def load_entry(loader, node):
    source = loader.construct_mapping(node, deep=True)
    if 'covert' in source:
        del source['covert']
    source['text'] = source.pop('task')
    for key in ('start', 'end'):
        if key in source:
            source[key] = source[key].astimezone(utctz)
    if 'comments' in source:
        source['comment'] = format_comments(source.pop('comments'))
    if 'post_result' in source:
        # Very old JIRA only ...
        data = source.pop('post_result')
        # The issue id is not in the post result. Take it from the entry text.
        key = JIRA_PATTERN.match(source['text']).group()
        submission = Submission(ext='jira', key=key, result=data)
        source['submissions'] = {submission.entry_index: submission}

    if 'post_results' in source:
        submissions = []
        for ext, things in source.pop('post_results').items():
            for key, data in things.items():
                submissions.append(Submission(ext=ext, key=key, result=data))
        source['submissions'] = {s.entry_index: s for s in submissions}
    return Entry(**source)


def load_sheet(loader, node):
    mapping = loader.construct_mapping(node, deep=True)
    if 'units' in mapping:
        return mapping['units']


class v0_Loader(yaml_Loader):
    pass

v0_Loader.add_constructor(u'tag:yaml.org,2002:timestamp', timestamp_constructor)
v0_Loader.add_constructor(u'!unit', load_entry)
v0_Loader.add_constructor(u'!sheet', load_sheet)
v0_Loader.add_constructor(u'tag:yaml.org,2002:python/object:impetuous.sheet.Sheet', load_sheet)


def load_legacy_sheet(doc):
    """
    I don't know what these parameters are, doc is probably a string or
    bytes or readable file object.

    Can return anything, but should return a list of entries ...
    """
    return yaml.load(doc, Loader=v0_Loader)



import datetime
import logging

import attr

from impetuous.ext import API, SECRET, LudicrousConditions, Submission

logger = logging.getLogger(__name__)


@attr.s(frozen=True)
class Freshdesk(API):

    IDENTIFIER = 'freshdesk'
    pattern = attr.ib()
    server = attr.ib()
    api_key = attr.ib(metadata={SECRET: True})

    def discover(self, impetuous, *entries):
        for entry in entries:
            time_spent = '%02i:%02i' % divmod(entry.duration_in_minutes, 60)
            for ticket_id in self.discover_by_pattern(entry, self.pattern):
                yield entry, Submission(ticket_id, {
                    'time_spent': time_spent,
                    'executed_at': entry.start.astimezone(datetime.timezone.utc).isoformat(),
                    'note': entry.comment,
                    'billable': '[billable]' in entry.comment,  # TODO consider using a tag/text pattern or something
                    'timer_running': not 'if I have anything to do with it',
                })

    async def agent(self, impetuous):
        return FreshdeskHttpAgent(self)


class FreshdeskHttpAgent(object):

    def __init__(self, api):
        import aiohttp
        self.api = api
        self.sess = aiohttp.ClientSession(auth=aiohttp.BasicAuth(self.api.api_key, 'X'))
        self.close = self.sess.close

    async def submit(self, sub):
        import aiohttp
        where = self.api.server + '/api/v2/tickets/{}/time_entries'.format(sub.key)
        resp = await self.sess.post(where, json=sub.data)
        try:
            resp.raise_for_status()
        except aiohttp.ClientResponseError as e:
            raise LudicrousConditions(e, "While submitting to {}: {}".format(resp.request_info.url, await resp.text()))
        else:
            return await resp.json()



# TODO rename to tables.py or something?
import logging
import os
import uuid
from datetime import datetime, timezone, time
from pathlib import Path

import attr
import yaml
from sqlalchemy import (DDL, CheckConstraint, Column, ForeignKey, Index,
                        Integer, MetaData, String, Table, UniqueConstraint,
                        create_engine, types)
from sqlalchemy.dialects.sqlite import INTEGER
from sqlalchemy.event import listen
from sqlalchemy.exc import DontWrapMixin
from sqlalchemy.sql.expression import join

DEFAULT_DATA_PATH = Path.home() / '.local' / 'share' / 'impetuous' / 'data.sqlite'
DATA_PATH = os.environ.get('IM_DB', DEFAULT_DATA_PATH)

logger = logging.getLogger(__name__)

try:
    from yaml import CSafeLoader as yaml_Loader
    from yaml import CSafeDumper as yaml_Dumper
except ImportError:
    logger.info("I couldn't import C yaml stuff so yaml will be slow.")
    from yaml import SafeLoader as yaml_Loader
    from yaml import SafeDumper as yaml_Dumper


def ensure_sane(engine):
    database = engine.url.database
    if database.lower() != ':memory:':  # What of other special values?
        dbdir = Path(database).parent
        try:
            dbdir.mkdir(parents=True)
        except FileExistsError:
            pass
        else:
            logger.info(_("I made %s for the database to live in because it didn't exist."), dbdir.parent)

    # Auto upgrade
    import alembic.config
    import alembic.command
    alembic_cfg = alembic.config.Config()
    alembic_cfg.set_main_option('script_location', 'impetuous:migrations')
    if (database.lower() != ':memory:' and Path(database).exists()):
        # Special case where the database kinda exists but the version isn't
        # set because early on we made the database before using alembic ...
        from alembic.migration import MigrationContext
        from alembic.autogenerate import compare_metadata
        mc = MigrationContext.configure(engine.connect())
        if mc.get_current_revision() is None:
            old_metadata = MetaData()
            Table(
                'entry', old_metadata,
                Column('id', UUID, default=uuid.uuid4, primary_key=True),
                Column('rev', UUID, default=uuid.uuid4, onupdate=uuid.uuid4, nullable=False),
                Column('text', String, nullable=False),
                Column('comment', String, nullable=False, default=''),
                Column('start', UTCDateTime, nullable=False),
                Column('end', UTCDateTime),
                Index('idx_entry_start', 'start', unique=True),
                CheckConstraint('end > start', name='ends after start'),
            )
            Table(
                'submission', old_metadata,
                Column('id', UUID, default=uuid.uuid4, primary_key=True),
                Column('rev', UUID, default=uuid.uuid4, onupdate=uuid.uuid4, nullable=False),
                Column('entry', UUID, ForeignKey('entry.id'), nullable=False),
                Column('ext', String, nullable=False),
                Column('key', String, nullable=False),
                Column('result', YamlDoc, nullable=False),
                Index('idx_submission_entry_ext', 'entry', 'ext', 'key', unique=True),
            )
            diff = compare_metadata(mc, old_metadata)
            if not diff:
                alembic.command.stamp(alembic_cfg, 'f4aaf1bdf8c0')
    alembic.command.upgrade(alembic_cfg, 'head')


class UUID(types.TypeDecorator):
    ''' UUID convenience type for SQLite.
    '''

    impl = types.BLOB

    def process_bind_param(self, value, dialect):
        return None if value is None else value.bytes

    def process_result_value(self, value, dialect):
        return None if value is None else uuid.UUID(bytes=value)


class InvalidDateTime(ValueError, DontWrapMixin):
    pass


class Time(types.TypeDecorator):
    impl = INTEGER

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        else:
            assert value.tzinfo is None
            assert value.microsecond == 0
            return value.hour * 3600 + value.minute * 60 + value.second

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        else:
            minute, second = divmod(value, 60)
            hour, minute = divmod(minute, 60)
            return time(hour=hour, minute=minute, second=second)


class UTCDateTime(types.TypeDecorator):
    ''' DateTime stored as UTC in database, timezone aware in application
    '''
    impl = INTEGER

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        elif hasattr(value, 'tzinfo'):  # duck typing???
            if value.tzinfo != timezone.utc:
                value = value.astimezone(timezone.utc)
            if value.microsecond != 0:
                raise InvalidDateTime("Seconds are the lowest resolution of time supported. This has %d microseconds." % value.microsecond)
            return value.timestamp()
        else:
            raise InvalidDateTime("Excepted a timezone-aware datetime, not a %s." % (type(value).__name__,))

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        else:
            # what about utcfromtimestamp? 🤔
            return datetime.fromtimestamp(value, tz=timezone.utc)


class YamlDoc(types.TypeDecorator):
    impl = types.String

    def process_bind_param(self, value, dialect):
        """ Return an object for inserting nito the database
        """
        if value is None:
            return None
        else:
            return yaml.dump(value, indent=2, default_flow_style=False, Dumper=yaml_Dumper)

    def process_result_value(self, value, dialect):
        """ Return a python object from a database value
        """
        if value is None:
            return None
        else:
            return yaml.load(value, Loader=yaml_Loader)


metadata = MetaData()

entry_t = Table(
    'entry', metadata,
    Column('id', UUID, default=uuid.uuid4, primary_key=True),
    # TODO this should auto update with a trigger or something ...
    Column('rev', UUID, default=uuid.uuid4, onupdate=uuid.uuid4, nullable=False),
    Column('text', String, nullable=False),
    Column('comment', String, nullable=False, default=''),
    Column('start', UTCDateTime, nullable=False),
    Column('end', UTCDateTime),
    Index('idx_entry_start', 'start', unique=True),
    Index('idx_entry_end', 'end'),
    # TODO store date times as big integers or floats or something...
    CheckConstraint('end > start', name='ends after start'),
)

submission_t = Table(
    'submission', metadata,
    Column('id', UUID, default=uuid.uuid4, primary_key=True),
    Column('rev', UUID, default=uuid.uuid4, onupdate=uuid.uuid4, nullable=False),
    # TODO read this http://docs.sqlalchemy.org/en/latest/dialects/sqlite.html#foreign-key-support
    Column('entry', UUID, ForeignKey('entry.id'), nullable=False),
    Column('ext', String, nullable=False),
    Column('key', String, nullable=False),
    Column('result', YamlDoc, nullable=False),
    Index('idx_submission_entry_ext', 'entry', 'ext', 'key', unique=True),
)

joins = {
    'entry submissions': join(entry_t, submission_t, entry_t.c.id == submission_t.c.entry),
    'submission entries': join(submission_t, entry_t, entry_t.c.id == submission_t.c.entry),
}

tables = {
    'entry': entry_t,
    'submission': submission_t,
}

SQLDIR = Path(__file__).parent / 'sql'


def read_ddl(name):
    return DDL((SQLDIR / name).open().read())


listen(entry_t, "after_create", read_ddl('check_insert_entry_dtrange.sql'))
listen(entry_t, "after_create", read_ddl('check_update_entry_dtrange.sql'))


#from sqlalchemy.event import listens_for
#from sqlalchemy.pool import Pool
#
#def meme():
#    print("MEME WAS CALLED")
#    return "hello"
#
#@listens_for(Pool, "connect")
#def my_on_connect(dbapi_con, connection_record):
#    c = dbapi_con
#    c.create_function("meme", 0, meme)
#    print("New DBAPI connection:", dbapi_con)



import logging
import re
from abc import ABC, abstractmethod, abstractproperty
from enum import Enum

import attr

from impetuous.config import CodedValue

logger = logging.getLogger(__name__)

SECRET = 'secret'  # Mark configuration values as secrets, for encoding/decoding


class LudicrousConditions(Exception):
    """
    Raise when when the module can't complete a request due invalid state like
    incomplete configuration or the weather or something. The program may
    display your error and quit or ignore it.
    """

    def __init__(self, error, suggestion):
        self.error = error
        self.suggestion = suggestion


class ConfigRequired(LudicrousConditions):

    @classmethod
    def from_attrs(cls, exc, other, section):
        setme = ", ".join(f"`{field.name}`" for field in attr.fields(other))
        return cls(exc, f"Maybe edit your config and set, {setme}, under the `[{section}]` section.")



class SubmissionStatus(Enum):

    invalid = 0
    unsubmitted = 1
    submitted = 2


class API(ABC):

    @abstractproperty
    def IDENTIFIER(self):
        """ A short string to pick this guy out of a crowd ...
        """

    @abstractmethod
    def discover(self, impetuous, *entries):
        """ yield (entry, Submission) pairs.
        """

    @abstractmethod
    async def agent(self, impetuous, ext):
        """
        Produce an object with a submit(Submission) coroutine method that can
        submit submissions to the given `ext`.
        """

    @classmethod
    def from_somewhere(cls, impetuous, section):
        return cls.from_config(impetuous.get_config_or_empty(), section)

    @classmethod
    def coded_fields(cls):
        return [field for field in attr.fields(cls) if field.metadata.get(SECRET)]

    @classmethod
    def from_config(cls, section):
        try:
            return cls(**{
                field.name:
                    CodedValue.decode_from_config(section, field.name)
                    if field.metadata.get(SECRET)
                    else section[field.name]
                for field in attr.fields(cls)
            })
        except KeyError as e:
            raise ConfigRequired.from_attrs('Key missing: %s' % e, cls, section.name)

    def discover_by_pattern(self, entry, pattern):
        logger.debug(_("%s looking for matches in %r using pattern %s."), type(self).__name__, entry.text, pattern)
        for match in re.findall(pattern, entry.text):
            logger.debug(_("%s found %r in %r using pattern %s."), type(self).__name__, match, entry.text, pattern)
            yield match


@attr.s(frozen=True)
class Submission(object):  # TODO this has the same name as impetuous.sheet.Submission
    """ A submission, maybe posted, maybe not, maybe not even postable.
    """
    # String to identify the submission for some extension in the db
    key = attr.ib()
    # Anything your heart desires ...
    data = attr.ib()
    # Short but readable label to identify the submission.
    label = attr.ib(default=attr.Factory(lambda self: self.key, takes_self=True))

    # TODO this should probably be an attribute as well ... almost positively
    # ... like absolutely, right? ... FIXME XXX
    def status(self, entry) -> SubmissionStatus:
        if entry.end is None:
            return SubmissionStatus.invalid
        elif entry.duration.total_seconds() <= 0:
            return SubmissionStatus.invalid
        elif entry.duration.total_seconds() < 60:
            logger.warning(_("%s has fewer than 60 seconds of time logged. JIRA/things will lose their minds I try to log this."), entry.text)
            return SubmissionStatus.invalid
        else:
            return SubmissionStatus.unsubmitted


@attr.s(frozen=True)
class Ext(object):
    """
    Modules allow for talking to external APIs so you can post time to JIRA or
    Freshdesk or whatever.

    #They shouldn't be too slow, as they may be created and run frequently, such
    #as whenever a entry is printed.
    """
    api = attr.ib()
    abbr = attr.ib()
    name = attr.ib()



import codecs
import collections
import configparser
import logging
import os
from pathlib import Path

logger = logging.getLogger(__name__)
CONFIG_DIR = Path.home() / '.config' / 'impetuous'
CONFIG_INI_PATH = CONFIG_DIR / 'config.ini'


class CodedValue(object):
    """ Some bytes encoded with codecs in that order.
    """

    def __init__(self, value: bytes, codecs: list):
        self.value = value
        self.codecs = codecs

    def __repr__(self):
        return "CodedValue(value={!r}, codecs={!r})".format(self.value, self.codecs)

    def __eq__(self, other):
        return (self.value == other.value
                and self.codecs == other.codecs)

    @property
    def is_encoded(self) -> bool:
        return bool(self.codecs)

    @property
    def encoded(self) -> str:
        """ Returns coded value as string.
        """
        return self.value.decode() # convert to string

    @property
    def decoded(self) -> str:
        """ Returns decoded value as string.
        """
        if self.codecs:
            return self.decode().decoded
        else:
            return self.value.decode('utf-8')

    @property
    def config_key_suffix(self):
        return '.'.join([''] + self.codecs)

    def decode(self):
        *codecs_, codec = self.codecs
        # Some of the binary codecs have characters that can't exist in a utf-8
        # string I guess, so in order to guarantee that a CodedValue can be
        # represented as a string, so that we can use it in a config file, we
        # base64 the output of every encoding (including base64 encoding)!
        return CodedValue(
                codecs.decode(codecs.decode(self.value, 'base64'), codec),
                codecs_)

    def encode(self, codec):
        return CodedValue(
                codecs.encode(codecs.encode(self.value, codec), 'base64'),
                self.codecs + [codec])

    @classmethod
    def plain(cls, value: str):
        return cls(value.encode(), [])

    @classmethod
    def from_config(cls, section, key):
        keys = [key_ for key_ in section.keys() if key_.startswith(key + '.')]
        if len(keys) > 1:
            raise KeyError("Ambiguous configuration, %r could be any of %r" % (key, keys))
        elif len(keys) == 1:
            key = keys[0]
            _, *codecs = key.split('.')
            return CodedValue(section[key].encode(), codecs)
        else:
            return CodedValue(section[key].encode(), [])

    @classmethod
    def decode_from_config(cls, section, key):
        return cls.from_config(section, key).decoded


def get_config(path=CONFIG_INI_PATH):
    config = configparser.ConfigParser()
    with path.open() as fp:
        config.readfp(fp)
    return config


# TODO use a safer write like how sheet files do it
def write_config(config, path=CONFIG_INI_PATH):
    with path.open('w') as fp:
        config.write(fp)


# TODO look at the bottom stuff and throw it out ...

#def fancy_extract_section_api(section):
#    section_apis = []
#
#    # This kind of rearranging and validating is lame but the config file
#    # format ends up looking nice at least?
#    for ident, cls in API_CLASSES.items():
#        api_data = {
#            key[len(ident)+1:]: value
#            for key, value in section.items()
#            if key.startswith(ident + '.')
#        }
#        if api_data:
#            section_apis.append(cls(**api_data))
#
#    if not section_apis:
#        logger.warning(_("No API information found for section `%s'"), section.name)
#    elif len(section_apis) > 1:
#        raise ValueError("Multiple APIs (%s) found for section `%s'; there can be only one!"
#                            % (", ".join(api.identifier for api in section_apis), section.name))
#    else:
#        return section_apis[0]
#
#
#def fancy_get_exts(config):
#    """ Get exts specified in the config
#    """
#    for __, section in config.items():
#        if section.name == 'DEFAULT':
#            continue
#        api = extract_section_api(section)
#        name = section.name
#        assert name, "Unnamed sections are for jerks"
#        yield Ext(api=api, name=name, abbr=section.get('abbr', name[0]),
#                  pattern=section['pattern'])



""" Serialization to load interaction constructs from YAML
"""
import logging
import re
from uuid import UUID

import dateutil.parser
import yaml

from impetuous.interaction import (Comparison, Conjunction, DeleteRequest,
                                   Field, FindRequest, Gather, InsertRequest,
                                   Param, UpdateRequest)

logger = logging.getLogger(__name__)


def deserialize_request(thing):
    return yaml.load(thing, Loader=APIYamlLoader)


def load_FindRequest(loader, node):
    return FindRequest(**loader.construct_mapping(node, deep=True))

def load_InsertRequest(loader, node):
    return InsertRequest(**loader.construct_mapping(node, deep=True))

def load_UpdateRequest(loader, node):
    return UpdateRequest(**loader.construct_mapping(node, deep=True))

def load_DeleteRequest(loader, node):
    return DeleteRequest(**loader.construct_mapping(node, deep=True))

def load_Gather(loader, node):
    return Gather(**loader.construct_mapping(node, deep=True))

def load_Comparison(loader, node):
    return Comparison(*loader.construct_sequence(node, deep=True))

def load_Conjunction_And(loader, node):
    parts = loader.construct_sequence(node, deep=True)
    return Conjunction.and_(parts=parts)

def load_Conjunction_Or(loader, node):
    parts = loader.construct_sequence(node, deep=True)
    return Conjunction.or_(parts=parts)

def load_Param(loader, node):
    return Param(loader.construct_scalar(node))

def load_Field(loader, node):
    return Field(**loader.construct_mapping(node, deep=True))

def load_datetime(loader, node):
    return dateutil.parser.parse(node.value)

def load_UUID(loader, node):
    return UUID(loader.construct_scalar(node))

class APIYamlLoader(yaml.SafeLoader):
    pass

APIYamlLoader.add_constructor('!im/find', load_FindRequest)
APIYamlLoader.add_constructor('!im/insert', load_InsertRequest)
APIYamlLoader.add_constructor('!im/update', load_UpdateRequest)
APIYamlLoader.add_constructor('!im/delete', load_DeleteRequest)
APIYamlLoader.add_constructor('!im/gather', load_Gather)
APIYamlLoader.add_constructor('!im/param', load_Param)
APIYamlLoader.add_constructor('!im/field', load_Field)
APIYamlLoader.add_constructor('!im/and', load_Conjunction_And)
APIYamlLoader.add_constructor('!im/or', load_Conjunction_Or)
APIYamlLoader.add_constructor('!im/cmp', load_Comparison)
APIYamlLoader.add_constructor('tag:yaml.org,2002:timestamp', load_datetime)
APIYamlLoader.add_constructor('!uuid', load_UUID)
UUID_PATTERN = re.compile(r'\d{8}-\d{4}-\d{4}-\d{4}-\d{12}')
APIYamlLoader.add_implicit_resolver('!uuid', UUID_PATTERN, None)



import asyncio
import itertools
import logging
import os
import subprocess
import tempfile
from datetime import timedelta, time
from random import choice, random, shuffle

import attr
import colorama as ca

from impetuous.cli import entry_fields, submission_fields
from impetuous.cli.display import Display
from impetuous.config import CONFIG_INI_PATH, CodedValue, write_config
from impetuous.ext import LudicrousConditions, SubmissionStatus
from impetuous.im import Impetuous
from impetuous.interaction import (Comparison, Conjunction, Field, FindRequest,
                                   Gather, OperationError, Param, Sort,
                                   TooFewResults, RequestError)
from impetuous.sheet import (Entry, EntryNotFoundError, Submission,
                             ValidationError, cli_yaml_Dumper, cli_yaml_Loader,
                             utcnow, localtz)

FUNCTION_DEST = 'the thing it is that you are trying to do'

logger = logging.getLogger(__name__)


match_entry_id = Comparison(Field('entry', 'id'), 'eq', Param('id'))
match_entry_since = Comparison(Field('entry', 'end'), 'ge', Param('since'))
match_entry_until = Comparison(Field('entry', 'start'), 'lt', Param('until'))

match_submission_id = Comparison(Field('submission', 'id'), 'eq', Param('id'))

# Doesn't end started before "until"
match_doesnt_end_and_started_before_until =\
    Conjunction.and_(
        Comparison(Field('entry', 'end'), 'is', None),
        Comparison(Field('entry', 'start'), 'lt', Param('until')),
    )

# Ends after "since" and starts before "until"
match_ends_after_since_and_starts_before_until =\
    Conjunction.and_(match_entry_since, match_entry_until)


class Agent(object):
    def is_authenticated(self):
        return True


cli_agent = Agent()


class DryRun(Exception):
    pass


def load_entry(interaction, match, using):  # TODO I don't really like this
    entry = interaction.find_one(
        Gather(fields=entry_fields, match=match, limit=1),
        using=using,
        unpack_items=Entry.joker,
    )
    submissions = interaction.find(
        Gather(fields=submission_fields, match=match_submission_id),
        using={"id": entry.id},
        unpack_items=Submission.joker,
    )
    entry.submissions = {s.entry_index: s for s in submissions}
    return entry


def load_entries(interaction, since, until, reverse=False, limit=None):
    """ Returns entries in chronological order, but you can reverse that.
    """
    entries_match = Conjunction.or_(
        match_doesnt_end_and_started_before_until,
        match_ends_after_since_and_starts_before_until,
    )
    order_by = (Sort('desc' if reverse else 'asc', Field('entry', 'start')),)

    entries = list(interaction.find(
        gather=Gather(fields=entry_fields,
                      match=entries_match,
                      limit=limit,
                      order_by=order_by),
        using={"since": since, "until": until},
        unpack_items=Entry.joker,
    ))

    submissions = interaction.find(
        gather=Gather(fields=submission_fields,
                      relating=('submission entries',),
                      match=entries_match and \
                              Comparison(Field('submission', 'entry'), 'in',
                                  [e.id for e in entries]),
                      order_by=order_by),
        using={"since": since, "until": until},
        unpack_items=Submission.joker,
    )
    # TODO, replace this thing around here and FindResult with an object that
    # will do both well ...
    submissions_by_entry_ids = {
        entry: tuple(submissions) for entry, submissions
        in itertools.groupby(submissions, lambda r: r.entry)
    }

    for entry in entries:
        if entry.id in submissions_by_entry_ids:
            submissions = submissions_by_entry_ids.pop(entry.id)
            entry.submissions = {post.entry_index: post for post in submissions}
        else:
            entry.submissions = {}
    return entries


def open_in_editor(path):
    if 'EDITOR' not in os.environ:
        logger.warning(_("EDITOR not set, using `ed`."))
        editor = 'ed'
    else:
        editor = '$EDITOR'
    subprocess.check_call(editor + ' ' + str(path), shell=True)



class Cli(Display):  # TODO, compose?

    def __init__(self, impetuous):
        super().__init__(impetuous)


async def im_repl(cli, args):
    from ptpython.repl import embed
    embed(globals(), locals())


async def im_a(_, args):
    what = ' '.join(getattr(args, '?'))
    article = getattr(args, FUNCTION_DEST)
    colors = [ca.Fore.GREEN, ca.Fore.BLUE, ca.Fore.MAGENTA]
    shuffle(colors)
    c = itertools.cycle(colors)
    if article == 'a' and what.lower() == 'real human being':
        response = "...", "and", "a", "real", "hero"
        punctuation = "!"
    else:
        response = "You", "are", article, *what.split()
        punctuation = "."
    def color():
        return next(c)
    for token in response[:-1]:
        print(token, color(), sep=" ", end="")
    print(response[-1], ca.Style.RESET_ALL, punctuation, ca.Style.RESET_ALL, sep="", end="\n")


async def im_show(cli, args):
    with cli.im.interact(agent=cli_agent) as i:
        entries = load_entries(i, args.since, args.until, reverse=args.reverse, limit=args.limit)
    cli.show(entries, verbose=args.verbose, reverse=args.reverse)


async def im_summary(cli, args):
    with cli.im.interact(agent=cli_agent) as i:
        entries = load_entries(i, args.since, args.until)

    def by_ext_name(i):
        return i[0].name
    def by_sub_key(i):
        return i[2].key
    def by_ext_name_sub_key(i):
        return i[0].name, i[2].key
    def tally(subs):
        return sum((entry.duration for ext, entry, sub in subs), timedelta())

    # iter_subs is: ext, entry, submission iterable
    iter_subs = cli.im.get_entry_submissions(*entries)
    subs = list(sorted(iter_subs, key=by_ext_name_sub_key))
    total = tally(subs)
    print(cli.format_duration(total), f"{ca.Style.BRIGHT}of time you'll never get back.{ca.Style.RESET_ALL}")
    for ext_name, ext_subs in itertools.groupby(subs, by_ext_name):
        ext_subs = list(ext_subs)
        total = tally(ext_subs)
        print(f"{ca.Style.BRIGHT}{ext_name}{ca.Style.RESET_ALL}", cli.format_duration(total))
        for sub_key, keyed_subs in itertools.groupby(ext_subs, by_sub_key):
            keyed_subs = list(keyed_subs)
            total = tally(keyed_subs)
            print(" ", f"{ca.Style.BRIGHT}{sub_key}{ca.Style.RESET_ALL}", cli.format_duration(total))
            for ext, entry, submission in keyed_subs:
                print("  ", cli.format_entry_start(entry),
                      cli.format_duration(entry.duration),
                      entry.text)


async def im_doing(cli, args):
    if args.blah:
        start_text = ' '.join(args.blah)
    else:
        start_text = None

    comment = args.comment
    # Simply round down to the second, never go ahead into the future ...
    when = args.when.replace(microsecond=0)

    with cli.im.interact(agent=cli_agent) as i:
        if start_text == '-':
            # Get the text of the last finished entry
            try:
                start_text, = i.find_one(
                        Gather(
                            fields=(Field('entry', 'text'),),
                            match=Conjunction.and_(
                                Comparison(Field('entry', 'end'), 'is not', None),
                                Comparison(Field('entry', 'start'), 'le', Param('until')),
                            ),
                            order_by=(Sort('desc', Field('entry', 'start')),),
                            limit=1,
                        ),
                        using={"until": when},
                        map_values=tuple,
                    )
            except TooFewResults:
                logger.error(_("No entry before %s to resume."), cli.format_datetime(when))
                raise SystemExit(1)
        # Try to finish the most recent running timer
        is_running_n_stuff = Conjunction.and_(
            Comparison(Field('entry', 'end'), 'is', None),
            Comparison(Field('entry', 'start'), 'le', Param('until')),
        )
        try:
            stop = i.find_one(
                Gather(
                    fields=(
                        Field('entry', 'id'),
                        Field('entry', 'rev'),
                        Field('entry', 'start'),
                        Field('entry', 'text')
                    ),
                    match=is_running_n_stuff,
                    order_by=(Sort('desc', Field('entry', 'start')),),
                    limit=1,
                ),
                using={"until": when},
                unpack_items=Entry.joker,
            )
        except TooFewResults:
            if start_text is None:
                print(_("Nothing running before %s. You are successfully continuing to do nothing...") % cli.format_datetime(when))
            else:
                # Nothing to stop, start something new
                entry = _start(i, when, start_text, comment)
                cli.print_entry(entry)
        else:
            if cli.im.settings.round_durations_to_minute:
                # If we're to stop this, the duration must be a multiple of a
                # minute. So move `when` back until a valid time given that.
                if stop.start.second > when.second:
                    when -= timedelta(minutes=1)
                when = when.replace(second=stop.start.second)
            if stop.start == when:
                # The current entry started when we want to be running, lets
                # rename it and use it instead
                if start_text is None:
                    print(_("Can't end %s at when it started. Maybe you want to delete it instead?") % cli.at_format(stop))
                    raise SystemExit(1)
                elif stop.text == start_text:
                    print(_("You're already doing %s") % cli.print_entry_text(stop, end=''))
                else:
                    print(_("Renaming"), cli.at_format(stop), "...")
                    values = {'text': start_text}
                    if comment is not None:
                        values['comment'] = comment
                    try:
                        i.update('entry', values, id=stop.id, rev=stop.rev)
                    except (RequestError, OperationError) as e:
                        logger.error(_("Failed to rename %s: %s"), cli.at_format(stop), e)
                    else:
                        entry = load_entry(i, match_entry_id, using={"id": stop.id})
                        cli.print_entry(entry)
            else:
                try:
                    i.update('entry', {'end': when}, id=stop.id, rev=stop.rev)
                except (RequestError, OperationError) as e:
                    logger.error(_("Failed to stop %s: %s"), cli.at_format(stop), e)
                else:
                    entry = load_entry(i, match_entry_id, using={"id": stop.id})
                    cli.print_entry(entry)
                # Finally start the thing we came here for ...
                if start_text is not None:
                    entry = _start(i, when, start_text, comment)
                    cli.print_entry(entry)
        if args.dry_run:
            raise DryRun()


def _start(i, when, text, comment):
    values = {"start": when, "text": text}
    if comment is not None:
        values['comment'] = comment
    inserted = i.insert('entry', values)
    return load_entry(i, match=match_entry_id, using=inserted)


async def im_suggest(cli, args):
    from impetuous.data import entry_t
    from sqlalchemy.sql import expression, func
    time_distance = func.abs(
            (utcnow().replace(microsecond=0) - entry_t.c.start + (12 * 3600))
            % (24 * 3600) - (12 * 3600)
            ).label('time_distance')
    date_distance = func.abs(
            (entry_t.c.start - utcnow().replace(microsecond=0)) / (24 * 3600)
            ).label('date_distance')
    distance = (time_distance / date_distance).label('distance')
    with cli.im.engine.begin() as conn:
        look_among = expression.select([entry_t.c.text, distance])\
            .order_by(entry_t.c.start.desc())\
            .limit(args.limit)
        q = expression.select([look_among.c.text])\
            .distinct(entry_t.c.text)\
            .order_by(look_among.c.distance.asc())
        for text, in conn.execute(q):
            print(text)


async def im_edit(cli, args):
    import yaml
    with cli.im.interact(agent=cli_agent) as i:
        entries = load_entries(i, args.since, args.until)

    data = yaml.dump(entries, indent=2, default_flow_style=False,
                     Dumper=cli_yaml_Dumper)
    with tempfile.NamedTemporaryFile(prefix='impetuous') as f:
        f.write(data.encode())
        f.flush()
        open_in_editor(f.name)
        f.seek(0)
        data = f.read()
    new_entries = yaml.load(data, Loader=cli_yaml_Loader)
    if new_entries is None:
        logger.error("Sheet file empty after editing, aborting! (Write an emtpy list with [] if you really want to delete your time entries.)")
        raise SystemExit(1)

    insertions = list()
    modifications = list()
    deletions = list()
    keyed = {entry.id: entry for entry in entries}
    for new in new_entries:
        if new.id is None:
            if new.submissions:
                raise NotImplementedError("Editing submissions is not supported.")
            insertions.append(new)
            logger.info("%s looks new", cli.at_format(new))
        else:
            try:
                old = keyed.pop(new.id)
            except KeyError:
                raise NotImplementedError("Did you put a new entry in here and set his primary key? You can't do that.")
            else:
                if old.submissions != new.submissions:
                    raise NotImplementedError("Editing submissions is not supported.")
                if old != new:
                    modifications.append(new)
    deletions = list(keyed.values())
    for entry in deletions:
        logger.info("%s looks deleted", cli.at_format(entry))

    with cli.im.interact(agent=cli_agent) as i:
        for entry in deletions:
            print("Deleting", cli.at_format(entry))
            i.delete('entry', id=entry.id, rev=entry.rev)
        for entry in modifications:
            print("Updating", cli.at_format(entry))
            i.update('entry', {
                "text": entry.text,
                "comment": entry.comment,
                "start": entry.start,
                "end": entry.end,
            }, id=entry.id, rev=entry.rev)
        for entry in insertions:
            print("Inserting", cli.at_format(entry))
            i.insert('entry', {
                "text": entry.text,
                "comment": entry.comment,
                "start": entry.start,
                "end": entry.end,
            })
        if args.dry_run:
            raise DryRun()


async def im_export(cli, args):
    raise NotImplementedError()


async def im_import(cli, args):
    raise NotImplementedError()


async def im_import_legacy(cli, args):
    from impetuous.legacy import load_legacy_sheet, v0_Loader
    entries = []
    logger.info(_("Loading entries..."))
    if v0_Loader.__base__.__name__ != 'CSafeLoader':
        logger.warning(_("Hey buddy, this loader %r is implemented in Python, not C, so it's very slow. For some reason I couldn't import yaml.CSafeLoader. If this is taking too long, go find your yaml.CSafeLoader and come back."), v0_Loader.__base__.__name__)
    for e, f in enumerate(args.files):
        if len(args.files) > 1:
            logger.info(_("Loading entries...") + " [%i/%i]", e, len(args.files))
        try:
            entries.extend(load_legacy_sheet(f))
        except ValueError as e:
            logger.error(_("Couldn't load %s because %s."), f.name, e)
    if entries:
        logger.info(_("Aligning entries..."))
        adjusted = timedelta()
        previous = None
        for entry in entries:
            if entry.start is None:
                logger.error(_("%s has no start time. The world is coming to an end."), cli.at_format(entry))
                raise NotImplementedError("wow")
            entry.start = entry.start.replace(microsecond=0)
            if entry.end is None:
                logger.warning(_("%s has no end so I've set the duration to zero and we'll skip over its."), cli.at_format(entry))
                entry.end = entry.start
                continue
            entry.end = entry.end.replace(microsecond=0)
            # Ensure that if the previous entry was made longer, it doesn't
            # overlap with this guy ...
            if previous is not None:
                if previous.end > entry.start:
                    delta = previous.end - entry.start
                    logger.debug(_("Advancing %s start by %s..."), cli.at_format(entry), delta)
                    entry.start += delta
                    adjusted += delta
                assert entry.start.microsecond == 0
            previous = entry
            # Round time entries to minute resolution depending on ...
            if cli.im.settings.round_durations_to_minute:
                excess = timedelta(seconds=entry.duration.total_seconds() % 60)
                assert excess.total_seconds() < 60
                if excess == 0:
                    pass
                elif excess.total_seconds() < 30 + adjusted.total_seconds():
                    # Shorten the duration by whatever excess it has ...
                    logger.debug(_("Reducing %s end by %s..."), cli.at_format(entry), excess)
                    entry.end -= excess
                    adjusted -= excess
                else:
                    # Add enough time to round to the 
                    change = timedelta(seconds=60) - excess
                    logger.debug(_("Advancing %s end by %s..."), cli.at_format(entry), change)
                    entry.end += change
                    adjusted += change
                assert entry.duration.total_seconds() % 60 == 0
        logger.info(_("Cumulative time adjustment is %s."), cli.format_duration(adjusted))
        # Attempt to insert entries entry ...
        logger.info(_("Importing entries; please hold the line while I write completely bogus data to your database that you didn't back up before running this."))
        failures = False
        with cli.im.interact(agent=cli_agent) as i:
            for entry in entries:
                if not entry.duration:
                    logger.warning(_("%s has no duration, I'm skipping it..."), cli.at_format(entry))
                    continue
                if entry.end < entry.start:
                    logger.warning(_("%s ends before it starts for some reason, I'm skipping it..."), cli.at_format(entry))
                    continue
                if not args.quiet:
                    cli.print_entry(entry, verbose=False)
                try:
                    inserted = i.insert('entry', {
                        "text": entry.text,
                        "comment": entry.comment,
                        "start": entry.start,
                        "end": entry.end,
                    })
                except (ValueError, OperationError) as e:
                    failures = True
                    logger.warning(_("Had a bad time trying to insert %s; %s"), cli.at_format(entry), e)
                else:
                    for submission in entry.submissions.values():
                        try:
                            i.insert('submission', {
                                "entry": inserted['id'],
                                "ext": submission.ext,
                                "key": submission.key,
                                "result": submission.result,
                            })
                        except (ValueError, OperationError) as e:
                            failures = True
                            logger.warning(_("Had a bad time trying to insert %s; %s"), cli.at_format(entry), e)
            if failures and not args.ignore_errors:
                logger.warning(_("Some data failed to import, so are rolling back. Rerun this with --ignore-errors if you don't care and wish to import anyway."))
                raise SystemExit(1)
            if args.dry_run:
                raise DryRun()
    else:
        logger.error(_("No entries found?"))


async def im_config_edit(cli, args):
    filepath = CONFIG_INI_PATH
    try:
        filepath.parent.mkdir(parents=True)
    except FileExistsError:
        pass
    else:
        logger.info(_("%s created."), filepath.parent)
    open_in_editor(filepath)


async def im_post(cli, args):
    with cli.im.interact(agent=cli_agent) as i:
        entries = load_entries(i, args.since, args.until)

    if not entries:
        logger.error(_("No entries found?"))
        raise SystemExit(1)

    agents = {}
    tasks = {}
    iter_subs = cli.im.get_entry_submissions(*entries)
    try:
        for ext, entry, submission in iter_subs:
            # TODO refactor this ...
            if (ext.name, submission.key) in entry.submissions:
                status = SubmissionStatus.submitted
            else:
                status = submission.status(entry)
            if status is SubmissionStatus.unsubmitted:
                if ext in agents:
                    agent = agents[ext]
                else:
                    agent = agents[ext] = await ext.api.agent(cli.im)
                job = PostJob(cli.im, ext, agent, entry, submission, args.dry_run)
                task = asyncio.get_event_loop().create_task(job.post())
                tasks[task] = job

        logger.debug("Posting jobs: %r", tasks)
        pending = tasks.keys()
        while pending:
            logger.debug(_("im_post waiting for: %r."), pending)
            done, pending = await asyncio.wait(pending, return_when=asyncio.FIRST_COMPLETED)
            for task in done:
                result = task.result()  # Rethrow exceptions
                job = tasks[task]
                if result is None:  # Assume an error in PostJob???
                    print("The Lord looked with favor on Abel and his offering, but on Cain and his offering (at ", end="")
                    cli.print_entry_start(job.entry, end=" Cain did ")
                    cli.print_entry_duration(job.entry, end=" of ")
                    cli.print_entry_submission(job.ext, job.entry, job.submission, verbose=True)
                    print(") he did not look with favor. So Cain was very angry, and his face was downcast.")
                else:
                    if not args.dry_run:
                        with cli.im.interact(agent=cli_agent) as i:
                            inserted = i.insert('submission', {
                                "entry": job.entry.id,
                                "ext": job.ext.name,
                                "key": job.submission.key,
                                "result": result,
                            })
                            submission = i.find_one(
                                gather=Gather(fields=submission_fields, match=match_submission_id),
                                using=inserted,
                                unpack_items=Submission.joker,
                            )
                            job.entry.submissions[submission.entry_index] = submission
                    print("And so it was written; at ", end="")
                    cli.print_entry_start(job.entry, end=" you did ")
                    cli.print_entry_duration(job.entry, end=" of ")
                    cli.print_entry_submission(job.ext, job.entry, job.submission, verbose=True)
                    print()
    finally:
        logger.debug(_("Cleaning up posting context."))
        for task in tasks.keys():
            if not task.done():
                logger.debug(_("Trying to tidy up incomplete %s."), task)
                task.cancel()
        if tasks:
            await asyncio.wait(tasks.keys())
        for agent in agents.values():
            await agent.close()


@attr.s(frozen=True, hash=False)
class PostJob(object):

    impetuous = attr.ib()
    ext = attr.ib()
    agent = attr.ib()
    entry = attr.ib()
    submission = attr.ib()
    dry_run = attr.ib()

    async def post(self):
        if self.dry_run:
            await asyncio.sleep(random() % 0.5)
            result = {"wow": "it worked!"}
        else:
            logger.debug(_("Submitting %r..."), self.submission)
            try:
                result = await self.agent.submit(self.submission)
            except LudicrousConditions as e:
                logger.error(_("%s failed to submit %r") + ': %s. %s',
                             type(self.ext).__name__, self.submission, e.error, e.suggestion)
                return
        logger.debug(_("Submitted %r with result %r."), self.submission, result)
        return result


def try_encode_field(section, field, codec):
    key = field.name
    try:
        current = CodedValue.from_config(section, key)
    except KeyError as err:
        logger.err(_("Something didn't work: %s"), err)
        raise SystemExit(1)
    else:
        logger.warning(_("Encoding `%s.%s%s` with %s."), section.name, key, current.config_key_suffix, codec)
        encoded = current.encode(codec)
        section[key + encoded.config_key_suffix] = encoded.encoded
        del section[key + current.config_key_suffix]


def try_decode_field(section, field):
    key = field.name
    try:
        current = CodedValue.from_config(section, key)
    except KeyError as err:
        logger.err(_("Something didn't work: %s"), err)
        raise SystemExit(1)
    else:
        if current.is_encoded:
            logger.warning(_("Decoding `%s.%s%s`."), section.name, key, current.config_key_suffix)
            decoded = current.decode()
            section[key + decoded.config_key_suffix] = decoded.encoded
            del section[key + current.config_key_suffix]
        else:
            logger.warning(_("`%s.%s%s` not encoded; doing nothing."), section.name, key, current.config_key_suffix)


async def im_encode(cli, args):
    config = cli.im.get_config_or_quit()
    for section, api in cli.im.section_apis(config):
        for field in api.coded_fields():
            try_encode_field(section, field, args.codec)
    write_config(config)


async def im_decode(cli, args):
    config = cli.im.get_config_or_quit()
    for section, api in cli.im.section_apis(config):
        for field in api.coded_fields():
            try_decode_field(section, field)
    write_config(config)


#async def im_insert(cli, args):
#    pass
#
#
#def im_insert_args(parser):
#    meme = {
#        'entry': entry_fields,
#    }
#    for mo
#    subparser = parser.add_parser('insert', formatter_class=ArgumentDefaultsHelpFormatter)
#    #subparser.set_defaults(action=im_insert)
#    subparser.required = True
#    subparser.dest = 'model'
#
#    for model, fields in meme.items():
#        subparser.
#    #a_args.add_argument('?', type=str, nargs='+')
#
#
#async def im_update(cli, args):
#    pass
#
#
#async def im_delete(cli, args):
#    pass



import logging
from contextlib import contextmanager
from datetime import timedelta

import attr
from sqlalchemy import create_engine

from impetuous.config import get_config
from impetuous.data import DATA_PATH, ensure_sane
from impetuous.ext import Ext, LudicrousConditions
from impetuous.ext.freshdesk import Freshdesk
from impetuous.ext.jira import Jira
from impetuous.interaction import Interaction, impetuous_access_control

logger = logging.getLogger(__name__)

NOT_EXT_SECTIONS = 'DEFAULT', 'impetuous'
API_CLASSES = {Jira.IDENTIFIER: Jira,
               Freshdesk.IDENTIFIER: Freshdesk}


@attr.s()
class Settings(object):
    round_durations_to_minute = attr.ib(default=True)


class Impetuous(object):
    """ Base program behaviour. TODO dissolve into nothing...?
    """
    # TODO cache most of these things? Or require explicit loading so state can
    # be invalidated explicitly.

    def __init__(self, dbecho=False):
        # TODO load config and settings from user or something? For one day
        # when this is a multi-user system ...
        try:
            self.config = get_config()
        except FileNotFoundError as e:
            logger.debug(_("Config could not be loaded: %r"), e)
            self.config = None
            self.config_error = e
        else:
            self.config_error = None
        self.settings = Settings()  # TODO load from impetuous section in config
        self.access_control = impetuous_access_control
        self.engine = create_engine(f'sqlite:///{DATA_PATH}', echo=dbecho)
        ensure_sane(self.engine)

    @contextmanager
    def interact(self, agent):
        with self.engine.connect() as conn:
            i = Interaction(conn, agent, self.access_control)
            with conn.begin():
                yield i

    def get_config_or_none(self):
        return self.config

    def get_config_or_empty(self):
        if self.config is None:
            return {}
        else:
            return self.config

    def get_config_or_quit(self):
        if self.config_error is not None:
            logger.error(self.config_error)
            logger.error(_("Try creating that file, see the readme for some ideas of what to put in it."))
            raise SystemExit(1)
        else:
            return self.config

    def get_entry_submissions(self, *entries):
        """
        Returns a list of module, entry, submission tuples.
        Ordered by module and entry for easy grouping.
        """
        # TODO don't call this so often and cache the dumb thing better ...
        for __, ext in self.section_exts(self.get_config_or_quit()):
            try:
                for entry, submission in ext.api.discover(self, *entries):
                    logger.debug(_("%s found submission %r."), ext.name, submission)
                    yield ext, entry, submission
            except LudicrousConditions as e:
                logger.warning(_("%s could not discover submissions") + ': %s. %s', ext.name, e.error, e.suggestion)

    def section_apis(self, config):
        """ Return section, API class pair.
        """
        for __, section in config.items():
            if section.name == 'DEFAULT':
                continue
            if 'api' in section:
                ident = section['api']
                if ident in API_CLASSES:
                    api = API_CLASSES[ident].from_config(section)
                    yield section, api
            else:
                logger.debug("Unrecognized section %s.", section)

    def section_exts(self, config):
        """ Returns section, Ext class pair.
        """
        for section, api in self.section_apis(config):
            name = section.get('name', section.name)
            assert name, "Unnamed sections are for jerks"
            abbr = section.get('abbr', name[0].upper())
            yield section, Ext(api=api, name=name, abbr=abbr)



"""empty message

Revision ID: ea11786324dd
Revises: f4aaf1bdf8c0
Create Date: 2018-03-17 00:37:08.218060

"""
from alembic import op
import sqlalchemy as sa


import impetuous.data

# revision identifiers, used by Alembic.
revision = 'ea11786324dd'
down_revision = 'f4aaf1bdf8c0'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index('idx_entry_end', 'entry', ['end'], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('idx_entry_end', table_name='entry')
    # ### end Alembic commands ###



import datetime
import logging
import textwrap

import colorama as ca

from impetuous.cli import get_terminal_width, maybe_elide, one_line
from impetuous.config import CONFIG_DIR, CONFIG_INI_PATH, CodedValue
from impetuous.ext import SubmissionStatus
from impetuous.sheet import Entry, ValidationError, localtz, utcnow

logger = logging.getLogger(__name__)


class Display(object):
    """ Mostly stuff about displaying to the CLI...

    Caches some things, so should be fairly shortish-lived.

    TODO most of the things in here like "print" should be format functions
    that return strings please. This is more useful and simplifies i18n.
    """

    def __init__(self, impetuous):
        self.im = impetuous
        self.local_today = utcnow().astimezone(localtz).date()
        self._terminal_width = None

    @property
    def terminal_width(self):
        if self._terminal_width is None:
            self._terminal_width = get_terminal_width()
        return self._terminal_width

    def _format_validation_err(self, err):
        params = err.params.copy()
        for key, value in params.items():
            if isinstance(value, Entry):
                entry = value
                params[key] = self.at_format(entry)
        return ValidationError(message=err.message, params=params)

    def at_format(self, entry):
        """ foo@420o'clock ...
        """
        return "".join((ca.Style.BRIGHT, entry.text, ca.Fore.BLUE, '@',
                        ca.Fore.GREEN, self.format_datetime(entry.start),
                        ca.Style.RESET_ALL))

    def show(self, entries, *, verbose=False, reverse=False):
        entries = sorted(entries, key=lambda entry: entry.start, reverse=reverse)
        while entries:
            entry = entries.pop(0)
            self.print_entry(entry, verbose=verbose)
            if entries:
                next_entry = entries[0]
                if reverse:
                    gap = entry.start - (next_entry.end or utcnow())
                else:
                    gap = next_entry.start - (entry.end or utcnow())
                if gap > datetime.timedelta(seconds=1):
                    print(ca.Fore.MAGENTA, '(', self.format_timedelta(gap), ')', ca.Style.RESET_ALL, sep='')

    def print_entry_submissions(self, entry, verbose=False):
        config = self.im.get_config_or_none()
        if config is not None:
            for ext, entry, submission in self.im.get_entry_submissions(entry):
                print(' ', end='')
                self.print_entry_submission(ext, entry, submission, verbose=verbose)

    def print_entry_submission(self, ext, entry, submission, *, verbose):
        if (ext.name, submission.key) in entry.submissions:
            status = SubmissionStatus.submitted
        else:
            status = submission.status(entry)
        if verbose:
            print(ext.name, end='')
            print('[', end='')
        else:
            print(ext.abbr, end='')
        if status is SubmissionStatus.invalid:
            print(ca.Style.BRIGHT, ca.Fore.RED, '!', ca.Style.RESET_ALL, sep='', end='')
        elif status is SubmissionStatus.unsubmitted:
            print(ca.Style.BRIGHT, '*', ca.Style.RESET_ALL, sep='', end='')
        if verbose:
            print(submission.label, end=']')

    def print_entry(self, entry, verbose=True):
        if verbose:
            self.print_entry_id(entry, end=' ')
        self.print_entry_start(entry, end=' ')
        self.print_entry_duration(entry, end=' ')
        self.print_entry_end(entry, end=' ')
        self.print_entry_text(entry, end='')
        self.print_entry_submissions(entry, verbose=verbose)
        if verbose:
            print()
            if entry.comment:
                print(textwrap.indent(entry.comment, '  '))
        else:
            if entry.comment:
                room = self.terminal_width - 28 - len(entry.text)  # start
                if entry.start.astimezone(localtz).date() != self.local_today:
                    room -= 11
                if entry.end.astimezone(localtz).date() != self.local_today:
                    room -= 11
                room -= 2 + 4 # surrounding braces and room for submissions?
                comment = maybe_elide(one_line(entry.comment), room)
                print(' (%s)' % comment, end='')
            print()

    def print_entry_id(self, entry, **kwargs):
        print(entry.id, **kwargs)

    def print_entry_text(self, entry, **kwargs):
        print(self.format_entry_text(entry), sep='', **kwargs)

    def print_entry_start(self, entry, **kwargs):
        print(self.format_entry_start(entry), sep='', **kwargs)

    def print_entry_end(self, entry, **kwargs):
        print(self.format_entry_end(entry), sep='', **kwargs)

    def format_entry_text(self, entry, **kwargs):
        return "".join((ca.Style.BRIGHT, entry.text, ca.Style.RESET_ALL))

    def format_entry_start(self, entry, **kwargs):
        return "".join((ca.Style.BRIGHT, ca.Fore.GREEN, self.format_datetime(entry.start), ca.Style.RESET_ALL))

    def format_entry_end(self, entry, **kwargs):
        return "".join((ca.Fore.GREEN, self.format_datetime(entry.end), ca.Fore.RESET))

    def format_datetime(self, dt):
        if dt is None:
            return '--:--:--'
        else:
            dt = dt.astimezone(localtz)
            is_today = dt.date() == self.local_today
            if is_today:
                return dt.time().strftime('%H:%M:%S')
            else:
                return dt.strftime('%Y-%m-%d %H:%M:%S')

    def print_entry_duration(self, entry, **kwargs):
        print(self.format_duration(entry.duration), **kwargs)

    def format_duration(self, delta, **kwargs):
        """ Some colours around format_timedelta()
        """
        return ''.join((ca.Fore.BLUE, ca.Style.BRIGHT, self.format_timedelta(delta), ca.Style.RESET_ALL))

    def format_timedelta(self, delta):
        seconds = int(delta.total_seconds())
        if seconds < 0:
            negate = True
            seconds = abs(seconds)
        else:
            negate = False
        minutes, seconds = divmod(seconds, 60)
        hours, minutes = divmod(minutes, 60)
        return '{}{:}:{:02}:{:02}'.format('-' if negate else '', hours, minutes, seconds)



"""
"""

# To use a consistent encoding
from codecs import open
from os import path

# Always prefer setuptools over distutils
from setuptools import find_packages, setup
from setuptools.command.develop import develop
from setuptools.command.install import install

# https://stackoverflow.com/questions/40051076/babel-compile-translation-files-when-calling-setup-py-install

def do_bable_stuff(distribution):
    from babel.messages.frontend import compile_catalog
    compiler = compile_catalog(distribution)
    option_dict = distribution.get_option_dict('compile_catalog')
    print(option_dict)
    compiler.domain = [option_dict['domain'][1]]
    compiler.directory = option_dict['directory'][1]
    compiler.run()


class InstallWithCompile(install):

    def run(self):
        do_bable_stuff(self.distribution)
        super().run()


class DevelopWithCompile(develop):

    def run(self):
        do_bable_stuff(self.distribution)
        super().run()


here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='impetuous',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=open('VERSION').read().strip(),

    description='CLI time tracking software indented for use with JIRA',
    long_description=long_description,

    # The project's main homepage.
    url='https://gitlab.com/sqwishy/impetuous',

    # Author details
    author='sqwishy',
    author_email='somebody@froghat.ca',

    # Choose your license
    license='GPLv3',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Office/Business',
        'Topic :: Utilities',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
    ],

    # What does your project relate to?
    keywords='time tracking',

    #cmdclass={
    #    'install': InstallWithCompile,
    #    'develop': DevelopWithCompile,
    #},

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=['impetuous', 'impetuous.cli', 'impetuous.ext',
              'impetuous.migrations', 'impetuous.migrations.versions'],
    #packages=find_packages(include=('impetuous/**',)),

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'colorama>=0.3.9',
        'python-dateutil>=2.6.0',
        'pyyaml',
        'pytz>=2018.3',
        'attrs>=17.4.0',
        'sqlalchemy>=1.2.0',
        'alembic>=0.9.8',
    ],

    setup_requires=['Babel'],
    tests_require=['pytest', 'pytest-benchmark'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'jira': ['aiohttp>=2.3.2'],
        'freshdesk': ['aiohttp>=2.3.2'],
        'test': ['pytest', 'pytest-benchmark'],
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={'': ['sql/*.sql']},
    #package_data={'': ['locale/*/*/*.mo', 'locale/*/*/*.po']},

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'im=impetuous.__main__:main',
        ],
    },
)



""" Impetuous and actions invoked from argument parser.
"""

import datetime
import logging
import re
import subprocess

import colorama as ca
import dateutil.parser

import impetuous.data
from impetuous.config import CONFIG_DIR, CONFIG_INI_PATH, write_config
from impetuous.interaction import Field
from impetuous.sheet import EntryNotFoundError, localtz, utcnow

logger = logging.getLogger(__name__)

entry_fields = [
    Field('entry', f) for f in impetuous.data.tables['entry'].c.keys()
]

submission_fields = [
    Field('submission', f) for f in impetuous.data.tables['submission'].c.keys()
]


def maybe_elide(string, max_length):
    r"""
    >>> maybe_elide('Hello world', 4)
    'H...'
    >>> maybe_elide('Hello world', 5)
    'H...d'
    >>> maybe_elide('Hello world', 9)
    'Hel...rld'
    >>> maybe_elide('Hello world', 10)
    'Hell...rld'
    >>> maybe_elide('Hello world', 11)
    'Hello world'
    >>> maybe_elide('Spam and eggs!', 9)
    'Spa...gs!'
    >>> maybe_elide('Spam and eggs!', 10)
    'Spam...gs!'
    >>> maybe_elide('We have phasers, I vote we blast \'em!   -- Bailey, "The Corbomite Maneuver", stardate 1514.2', 29)
    'We have phase...ardate 1514.2'

    """
    assert max_length > 3
    if len(string) > max_length:
        chunklen = (max_length - 3) // 2
        return "{}...{}".format(string[:chunklen + 1],
                                '' if chunklen == 0 else string[-chunklen:])
    else:
        return string


def one_line(string):
    r"""
    Sanitize for one-line printing?

    >>> sanitize('foo\nbar\tbaz')
    'foo bar baz'

    """
    return string.replace('\t', ' ').replace('\n', ' ')


def get_terminal_width():
    """
    Return the terminal width as an integer or None if we can't figure it out.
    """
    try:
        stty = subprocess.check_output(['stty', 'size'])
    except (FileNotFoundError, subprocess.CalledProcessError):
        pass
    else:
        try:
            return int(stty.strip().split(b' ')[1])
        except (ValueError, IndexError):
            pass
    try:
        tput = subprocess.check_output(['tput', 'cols'])
    except (FileNotFoundError, subprocess.CalledProcessError):
        pass
    else:
        try:
            return int(tput.strip())
        except ValueError:
            pass
    try:
        env = subprocess.check_output('echo -n $COLUMNS', shell=True)
    except (FileNotFoundError, subprocess.CalledProcessError):
        pass
    else:
        try:
            return int(env)
        except ValueError:
            pass


def parse_friendly_datetime(value, default=utcnow().astimezone(localtz), **kwargs):
    """
    Only really seems to know what UTC and your local time zone is, everything
    else it'll ignore and we pretend like it's your local timezone ...

    >>> parse_friendly_datetime("1-1-2008 13:00:09")
    datetime.datetime(2008, 1, 1, 13, 0, 9, tzinfo=tzlocal())
    """
    if value == 'now':
        return default
    else:
        dt = dateutil.parser.parse(value, **kwargs)
        if dt.tzinfo is None:
            return dt.replace(tzinfo=localtz)
        else:
            return dt


class WhenTime(object):

    class EntrySeek(object):

        class StartEdge:
            pass

        class EndEdge:
            pass

        def __init__(self, count, edge):
            self.count = count
            self.edge = edge

    def __init__(self, initial, seek=None, delta=None):
        assert initial.tzinfo is not None
        self.initial = initial
        self.seek = seek
        self.delta = delta

    def absolute_datetime(self, entries):
        """
        Return an absolute date and time, possibly relative to some given
        entries if this WhenType specifies a delta (such as relative to a entry's
        start/end time).
        """
        value = self.initial
        if self.seek is not None:
            steps = []
            for entry in entries:
                if self.seek.edge is self.seek.StartEdge:
                    time = entry.start
                else:
                    time = entry.end
                if time is None:
                    raise ValueError(_("Refusing seek past null time."))
                elif time < value:
                    steps.append(time)
                else:
                    break
            try:
                value = steps[-self.seek.count]
            except IndexError:
                raise EntryNotFoundError(_("No entry matching seek."))
        if self.delta is not None:
            value += self.delta
        return value


class WhenArgType(object):
    """
    It's important to use utcnow and the local timezone as attributes on this
    guy instead of just localtimenow because otherwise, time zones will fuck
    you in weird ways.

    Trust me. UTC is the only sane time zone. Use it as much as possible.
    """

    when_format_re = re.compile(
        # Fixed initial
        r'((\.|now)|(?:(\d{1,2}):(\d{1,2}):?(\d{1,2})?))?'
        # Seek
        r'(([[\]])(\d*))?'
        # Relative offset
        r'(([+-])(\d+h)?(\d+m)?(\d+s)?)?'
        r'$'
    )

    def __init__(self, utcnow=utcnow, localtz=localtz):
        self.utcnow = utcnow
        self.localtz = localtz

    def __repr__(self):
        """ For formatting nicely when displayed in an argument parsing message.
        """
        return 'WhenArgType'

    def __call__(self, text):
        """
        Returns a WhenTime which can be used to produce an absolute time given
        the context of a collection of entries.
        """
        if text == 'now':
            return WhenTime(self.utcnow())

        match = self.when_format_re.match(text)
        if match is not None:
            (initial, initial_dot, initial_hour, initial_minute, initial_second,
             seek, seek_edge, seek_count,
             delta, delta_sign, delta_hours, delta_minutes, delta_seconds
             ) = match.groups()

            # Set initial
            if initial is None or initial_dot:
                initial = self.utcnow()
            else:
                time = datetime.time(hour=int(initial_hour),
                                     minute=int(initial_minute),
                                     second=0 if initial_second is None else int(initial_second[0:]))
                now = self.utcnow()
                today = now.astimezone(self.localtz).date()
                initial = datetime.datetime.combine(today, time).replace(tzinfo=self.localtz)

            # Set seek
            if seek is not None:
                if seek_edge == '[':
                    seek_edge = WhenTime.EntrySeek.StartEdge
                else:
                    seek_edge = WhenTime.EntrySeek.EndEdge
                seek = WhenTime.EntrySeek(edge=seek_edge,
                                         count=int(seek_count) if seek_count else 1)

            # Finally make relative/delta time adjustment
            if delta is not None:
                delta = datetime.timedelta(
                    hours=0 if delta_hours is None else int(delta_hours[:-1]),
                    minutes=0 if delta_minutes is None else int(delta_minutes[:-1]),
                    seconds=0 if delta_seconds is None else int(delta_seconds[:-1])
                )
                if delta_sign == '-':
                    delta = -delta

            return WhenTime(initial=initial, seek=seek, delta=delta)
        else:
            raise ValueError(_("Bad format '{}'.").format(text[0:]))



import asyncio
import logging

import attr

from impetuous.ext import API, SECRET, LudicrousConditions, Submission

logger = logging.getLogger(__name__)


@attr.s(frozen=True)
class Jira(API):

    IDENTIFIER = 'jira'
    pattern = attr.ib()
    server = attr.ib()
    basic_auth = attr.ib(metadata={SECRET: True},
                         convert=lambda value: tuple(value.split(':', 1)))

    def discover(self, impetuous, *entries):
        for entry in entries:
            for issuekey in self.discover_by_pattern(entry, self.pattern):
                yield entry, Submission(issuekey, {
                    'started': entry.start.strftime('%Y-%m-%dT%H:%M:%S.000%z'),
                    'timeSpentSeconds': entry.duration_in_minutes * 60,
                    'comment': entry.comment,
                })

    async def agent(self, impetuous):
        return JiraHttpAgent(self)


class JiraHttpAgent(object):

    def __init__(self, api):
        import aiohttp
        self.api = api
        self.sess = aiohttp.ClientSession(auth=aiohttp.BasicAuth(*api.basic_auth))
        self.close = self.sess.close

    async def submit(self, sub):
        import aiohttp
        where = self.api.server + '/rest/api/2/issue/{}/worklog'.format(sub.key)
        resp = await self.sess.post(where, json=sub.data, params={'notifyUsers': 'false'})
        try:
            resp.raise_for_status()
        except aiohttp.ClientResponseError as e:
            raise LudicrousConditions(e, "While submitting to {}: {}".format(resp.request_info.url, await resp.text()))
        else:
            return await resp.json()
